@@ -8,10 +8,10 @@ _logger = logging.getLogger("OnboardGtk")
 import sys
 import time
 import traceback
-import gobject
-gobject.threads_init()
+import signal
+
+from gi.repository import GObject, Gdk, Gtk
 
-import gtk
 import virtkey
 import gettext
 import os.path
@@ -21,7 +21,6 @@ from gettext import gettext as _
 from Onboard.Indicator import Indicator
 from Onboard.Keyboard import Keyboard
 from Onboard.KeyGtk import *
-from Onboard.Pane import Pane
 from Onboard.KbdWindow import KbdWindow, KbdPlugWindow
 from Onboard.KeyboardSVG import KeyboardSVG
 from Onboard.utils       import show_confirmation_dialog, CallOnce
@@ -43,6 +42,7 @@ gettext.bindtextdomain(app)
 
 DEFAULT_FONTSIZE = 10
 
+
 class OnboardGtk(object):
     """
     This class is a mishmash of things that I didn't have time to refactor in to seperate classes.
@@ -56,14 +56,30 @@ class OnboardGtk(object):
     """ The keyboard widget """
     keyboard = None
 
+    restart = False
+
     def __init__(self, main=True):
         sys.path.append(os.path.join(config.install_dir, 'scripts'))
 
+        self.init()
+
+        if main:
+            # Release enter key when killing onboard by
+            # pressing killall onboard in the console.
+            # -> Disabled: This gets onboard stuck on exit in
+            # gnome-screensaver until the pointer moves over the keyboard.
+            # May be a GTK bug, disabled for now (Oneiric).
+            #signal.signal(signal.SIGTERM, self.on_signal)
+
+            _logger.info("Entering mainloop of onboard")
+            Gtk.main()
+
+    def init(self):
         self.keyboard_state = None
         self.vk_timer = None
         self.reset_vk()
+        self._connections = []
 
-        # create main window
         if config.xid_mode:    # XEmbed mode for gnome-screensaver?
             self._window = KbdPlugWindow()
 
@@ -72,42 +88,68 @@ class OnboardGtk(object):
             sys.stdout.flush()
         else:
             self._window = KbdWindow()
-            self._window.connect_object("quit-onboard",
-                                        self.do_quit_onboard, None)
-
-        _logger.info("Getting user settings")
+            self.do_connect(self._window, "quit-onboard",
+                            lambda x: self.do_quit_onboard())
+        self._window.application = self
 
         # load the initial layout
-        self.update_layout()
+        _logger.info("Loading initial layout")
+        self.reload_layout()
 
         # connect notifications for keyboard map and group changes
-        self.keymap = gtk.gdk.keymap_get_default()
-        self.keymap.connect("keys-changed", self.cb_keys_changed) # map changes
-        gtk.gdk.event_handler_set(cb_any_event, self)          # group changes
+        self.keymap = Gdk.Keymap.get_default()
+        self.do_connect(self.keymap, "keys-changed", self.cb_keys_changed) # map changes
+        Gdk.event_handler_set(cb_any_event, self)          # group changes
 
         # connect config notifications here to keep config from holding
         # references to keyboard objects.
-        once = CallOnce(20).enqueue  # delay callbacks 50ms
-        theme_change  = lambda x: once(self.cb_theme_changed, x)
-        update_layout = lambda x: once(self.update_layout, True)
-        queue_draw    = lambda x: once(self.keyboard.queue_draw)
+        once = CallOnce(50).enqueue  # delay callbacks by 50ms
+        reload_layout = lambda x: once(self.reload_layout, True)
+        update_ui     = lambda x: once(self.update_ui)
+        redraw        = lambda x: once(self.keyboard.redraw)
 
-        config.layout_filename_notify_add(update_layout)
-        config.color_scheme_filename_notify_add(update_layout)
-        config.key_label_overrides_notify_add(update_layout)
-        config.theme_attributes_notify_add(queue_draw)
-        config.snippets_notify_add(update_layout)
+        config.layout_filename_notify_add(reload_layout)
+        config.key_label_font_notify_add(reload_layout)
+        config.key_label_overrides_notify_add(reload_layout)
 
-        config.scanning_notify_add(lambda x: \
+        config.theme_settings.color_scheme_filename_notify_add(reload_layout)
+        config.theme_settings.key_label_font_notify_add(reload_layout)
+        config.theme_settings.key_label_overrides_notify_add(reload_layout)
+        config.theme_settings.theme_attributes_notify_add(redraw)
+
+        config.snippets_notify_add(reload_layout)
+
+        config.show_click_buttons_notify_add(update_ui)
+        if config.mousetweaks:
+            config.mousetweaks.state_notify_add(update_ui)
+
+        config.window_decoration_notify_add(self._cb_recreate_window)
+        config.force_to_top_notify_add(self._cb_recreate_window)
+
+        config.transparency_notify_add( \
+                        lambda x: self.keyboard.update_transparency())
+        config.background_transparency_notify_add(redraw)
+        config.transparent_background_notify_add(update_ui)
+        config.enable_inactive_transparency_notify_add( \
+                        lambda x: self.keyboard.update_transparency())
+        config.inactive_transparency_notify_add( \
+                        lambda x: self.keyboard.update_inactive_transparency())
+
+        config.enable_scanning_notify_add(lambda x: \
                                      self.keyboard.reset_scan())
 
+
         # create status icon
-        self.status_icon = Indicator(self._window)
-        self.status_icon.connect("quit-onboard", self.do_quit_onboard)
+        # Indicator is a singleton to allow recreating the keyboard
+        # window on changes to the "force_to_top" setting.
+        self.status_icon = Indicator()
+        self.status_icon.set_keyboard_window(self._window)
+        self.do_connect(self.status_icon, "quit-onboard",
+                        lambda x: self.do_quit_onboard())
 
         # Callbacks to use when icp or status icon is toggled
         config.show_status_icon_notify_add(self.show_hide_status_icon)
-        config.icp_in_use_change_notify_add(self.cb_icp_in_use_toggled)
+        config.icp.in_use_notify_add(self.cb_icp_in_use_toggled)
 
         self.show_hide_status_icon(config.show_status_icon)
 
@@ -116,16 +158,29 @@ class OnboardGtk(object):
 
         # Minimize to IconPalette if running under GDM
         if os.environ.has_key('RUNNING_UNDER_GDM'):
-            config.icp_in_use = True
+            config.icp.in_use = True
             config.show_status_icon = False
             self.show_hide_taskbar()
 
+        # Check gnome-screen-saver integration
+        # onboard_xembed_enabled                False True     True      True
+        # config.gss.embedded_keyboard_enabled  any   False    any       False
+        # config.gss.embedded_keyboard_command  any   empty    !=onboard ==onboard
+        # Action:                               nop   enable   Question1 Question2
+        #                                             silently
+        if not config.xid_mode and \
+           config.onboard_xembed_enabled:
 
-        # If onboard is configured to be embedded into the unlock screen
-        # dialog, and the embedding command is not set to onboard, ask
-        # the user what to do
-        if config.onboard_xembed_enabled:
-            if not config.is_onboard_in_xembed_command_string():
+            # If it appears, that nothing has touched the gss keys before,
+            # silently enable gss integration with onboard.
+            if not config.gss.embedded_keyboard_enabled and \
+               not config.gss.embedded_keyboard_command:
+                config.enable_gss_embedding(True)
+
+            # If onboard is configured to be embedded into the unlock screen
+            # dialog, and the embedding command is different from onboard, ask
+            # the user what to do
+            elif not config.is_onboard_in_xembed_command_string():
                 question = _("Onboard is configured to appear with the dialog to "
                              "unlock the screen; for example to dismiss the "
                              "password-protected screensaver.\n\n"
@@ -137,13 +192,11 @@ class OnboardGtk(object):
                              "Onboard when unlocking the screen?")
                 reply = show_confirmation_dialog(question)
                 if reply == True:
-                    config.onboard_xembed_enabled = True
-                    config.gss_xembed_enabled = True
-                    config.set_xembed_command_string_to_onboard()
+                    config.enable_gss_embedding(True)
                 else:
                     config.onboard_xembed_enabled = False
             else:
-                if not config.gss_xembed_enabled:
+                if not config.gss.embedded_keyboard_enabled:
                     question = _("Onboard is configured to appear with the dialog "
                                  "to unlock the screen; for example to dismiss "
                                  "the password-protected screensaver.\n\n"
@@ -151,18 +204,20 @@ class OnboardGtk(object):
                                  "Would you like to activate it?")
                     reply = show_confirmation_dialog(question)
                     if reply == True:
-                        config.onboard_xembed_enabled = True
-                        config.gss_xembed_enabled = True
-                        config.set_xembed_command_string_to_onboard()
+                        config.enable_gss_embedding(True)
                     else:
                         config.onboard_xembed_enabled = False
 
 
-        if main:
-            _logger.info("Entering mainloop of onboard")
-            gtk.main()
-            self.clean()
+    def do_connect(self, instance, signal, handler):
+        handler_id = instance.connect(signal, handler)
+        self._connections.append((instance, handler_id))
 
+    def on_signal(self, signum, frame):
+        if signum == signal.SIGTERM:
+            _logger.debug("SIGTERM received")
+            self.cleanup()
+            sys.exit(1)
 
     # Method concerning the taskbar
     def show_hide_taskbar(self):
@@ -172,7 +227,7 @@ class OnboardGtk(object):
         This method should be called every time such an alternative way
         is activated or deactivated.
         """
-        if config.icp_in_use or \
+        if config.icp.in_use or \
            config.show_status_icon:
             self._window.set_property('skip-taskbar-hint', True)
         else:
@@ -183,19 +238,19 @@ class OnboardGtk(object):
     def cb_icp_in_use_toggled(self, icp_in_use):
         """
         This is the callback that gets executed when the user toggles
-        the gconf key named in_use of the icon_palette. It also
+        the gsettings key named in_use of the icon_palette. It also
         handles the showing/hiding of the taskar.
         """
         _logger.debug("Entered in on_icp_in_use_toggled")
         if icp_in_use:
             # Show icon palette if appropriate and handle visibility of taskbar.
-            if self._window.hidden:
-                self._window.icp.do_show()
+            if not self._window.is_visible():
+                self._window.icp.show()
             self.show_hide_taskbar()
         else:
             # Show icon palette if appropriate and handle visibility of taskbar.
-            if self._window.hidden:
-                self._window.icp.do_hide()
+            if not self._window.is_visible():
+                self._window.icp.hide()
             self.show_hide_taskbar()
         _logger.debug("Leaving on_icp_in_use_toggled")
 
@@ -203,7 +258,7 @@ class OnboardGtk(object):
     # Methods concerning the status icon
     def show_hide_status_icon(self, show_status_icon):
         """
-        Callback called when gconf detects that the gconf key specifying
+        Callback called when gsettings detects that the gsettings key specifying
         whether the status icon should be shown or not is changed. It also
         handles the showing/hiding of the taskar.
         """
@@ -221,32 +276,29 @@ class OnboardGtk(object):
 
         TODO would be nice if appeared to iconify to taskbar
         """
-        if self._window.hidden: self._window.deiconify()
-        else: self._window.iconify()
+        self._window.toggle_visible()
 
 
-    # Methods concerning the listening to keyboard layout changes
+    # keyboard layout changes
     def cb_keys_changed(self, *args):
-        self.update_layout()
+        self.reload_layout()
 
     def cb_vk_timer(self):
         """
         Timer callback for polling until virtkey becomes valid.
         """
         if self.get_vk():
-            self.update_layout(force_update=True)
-            gobject.source_remove(self.vk_timer)
+            self.reload_layout(force_update=True)
+            GObject.source_remove(self.vk_timer)
             self.vk_timer = None
             return False
         return True
 
-    def cb_theme_changed(self, theme_filename):
-        # load and apply the theme
-        theme = Theme.load(theme_filename)
-        if theme:
-            theme.apply()
+    def update_ui(self):
+        self.keyboard.update_ui()
+        self.keyboard.redraw()
 
-    def update_layout(self, force_update=False):
+    def reload_layout(self, force_update=False):
         """
         Checks if the X keyboard layout has changed and
         (re)loads Onboards layout accordingly.
@@ -268,21 +320,21 @@ class OnboardGtk(object):
 
         if self.keyboard_state != keyboard_state or force_update:
             self.keyboard_state = keyboard_state
-            self.load_layout(config.layout_filename, 
-                             config.color_scheme_filename)
+            self.load_layout(config.layout_filename,
+                             config.theme_settings.color_scheme_filename)
 
-        # if there is no X keyboard, poll until it appears
+        # if there is no X keyboard, poll until it appears (if ever)
         if not vk and not self.vk_timer:
-            self.vk_timer = gobject.timeout_add_seconds(1, self.cb_vk_timer)
+            self.vk_timer = GObject.timeout_add_seconds(1, self.cb_vk_timer)
 
     def load_layout(self, layout_filename, color_scheme_filename):
         _logger.info("Loading keyboard layout from " + layout_filename)
         if (color_scheme_filename):
             _logger.info("Loading color scheme from " + color_scheme_filename)
         if self.keyboard:
-            self.keyboard.clean()
-        self.keyboard = KeyboardSVG(self.get_vk(), 
-                                    layout_filename, 
+            self.keyboard.cleanup()
+        self.keyboard = KeyboardSVG(self.get_vk(),
+                                    layout_filename,
                                     color_scheme_filename)
         self._window.set_keyboard(self.keyboard)
 
@@ -306,26 +358,63 @@ class OnboardGtk(object):
 
 
     # Methods concerning the application
-    def clean(self):
-        self.keyboard.clean()
-        self._window.hide()
-
-    def quit(self, widget=None):
-        self.clean()
-        gtk.main_quit()
-
-    def do_quit_onboard(self, data=None):
+    def do_quit_onboard(self, restart = False):
         _logger.debug("Entered do_quit_onboard")
-        self._window.save_size_and_position()
-        gtk.main_quit()
+        self.restart = restart
+        if not restart:
+            self.final_cleanup()
+        self.cleanup()
+        Gtk.main_quit()
 
+    def cleanup(self):
+        if not config.xid_mode:
+            self._window.save_size_and_position()
+        config.cleanup()
+
+        # Make an effort to disconnect all handlers. This may
+        # still not be enough to remove all references to windows
+        # but it ought to reduce the chances of side-effects when
+        # restarting due to changes to the window type hint.
+        for instance, handler_id in self._connections:
+            instance.disconnect(handler_id)
+
+        if self.keyboard:
+            self.keyboard.cleanup()
+            self._window.keyboard.destroy()  # necessary?
+        self.status_icon.set_keyboard_window(None)
+        self._window.icp.hide()   # save position
+        self._window.destroy()
+        self._window = None
+
+    def final_cleanup(self):
+        config.final_cleanup()
+
+    def _cb_recreate_window(self, value):
+        # Window type hint can only be set on window creation.
+        # Same on gnome-shell for window decoration.
+        # -> force restart
+        self.do_quit_onboard(restart=True)
 
 def cb_any_event(event, onboard):
     # Update layout on keyboard group changes
-    # XkbStateNotify maps to gtk.gdk.NOTHING
+    # XkbStateNotify maps to Gdk.EventType.NOTHING
     # https://bugzilla.gnome.org/show_bug.cgi?id=156948
-    if event.type == gtk.gdk.NOTHING:
-        onboard.update_layout()
-    gtk.main_do_event(event)
 
+    # Hide bug in Oneirics GTK3
+    # Suppress ValueError: invalid enum value: 4294967295
+    type = None
+    try:
+        type = event.type
+    except ValueError:
+        pass
+
+    if 0: # debug
+        print event, event.type
+        if type == Gdk.EventType.VISIBILITY_NOTIFY:
+            print event.state
+
+    if type == Gdk.EventType.NOTHING:
+        onboard.reload_layout()
+
+    Gtk.main_do_event(event)
 
